<?php
/*********************************************
 *  CPG Dragonfly™ CMS
 *********************************************
	Copyright © since 2010 by CPG-Nuke Dev Team
	http://dragonflycms.org

	Dragonfly is released under the terms and conditions
	of the GNU GPL version 2 or any later version
*/

namespace Poodle;

abstract class Hash
{
	const
		BCRYPT_DEFAULT_WORK_FACTOR = 10;

	public static
		$PBKDF2_HASH_ALGORITHM = 'sha256',
		$PBKDF2_ITERATIONS = 1000,
		$PBKDF2_SALT_BYTES = 24,
		$PBKDF2_HASH_BYTES = 24,

		$SCRYPT_SALT_BYTES = 16,
		$SCRYPT_HASH_BYTES = 32,
		$SCRYPT_CPU_COST   = 1024, // 16384 = slow
		$SCRYPT_BLOCK_SIZE = 8,
		$SCRYPT_ITERATIONS = 1;

	public static function algos()
	{
		return array('bcrypt','pbkdf2','scrypt') + hash_algos();
	}

	public static function available($algo)
	{
		return (in_array($algo, self::algos()) || function_exists($algo));
	}

	public static function file($algo, $filename, $raw=false)
	{
		if (in_array($algo, hash_algos())) {
			return hash_file($algo, $filename, $raw);
		}
		$algo = $algo.'_file';
		if (function_exists($algo)) {
			return $algo($filename, $raw);
		}
		return false;
	}

	public static function string($algo, $string, $raw=false)
	{
		switch ($algo)
		{
		case 'none':
		case '':
			return $string;

		case 'bcrypt':
			return self::bcrypt($string);

		case 'pbkdf2':
			// format: algorithm$iterations$salt$hash
			$salt = base64_encode(\Poodle\Random::bytes(static::$PBKDF2_SALT_BYTES));
			return static::$PBKDF2_HASH_ALGORITHM
				.'$'.static::$PBKDF2_ITERATIONS
				.'$'.$salt
				.'$'.base64_encode(static::pbkdf2(
					static::$PBKDF2_HASH_ALGORITHM,
					$string,
					$salt,
					static::$PBKDF2_ITERATIONS,
					static::$PBKDF2_HASH_BYTES,
					true
				));

		case 'scrypt':
			$salt = base64_encode(\Poodle\Random::bytes(static::$SCRYPT_SALT_BYTES));
			return static::$SCRYPT_CPU_COST
				.'$'.static::$SCRYPT_BLOCK_SIZE
				.'$'.static::$SCRYPT_ITERATIONS
				.'$'.$salt
				.'$'.base64_encode(Scrypt::hash($string, $salt,
					static::$SCRYPT_CPU_COST,
					static::$SCRYPT_BLOCK_SIZE,
					static::$SCRYPT_ITERATIONS,
					static::$SCRYPT_HASH_BYTES,
					true));

		default: # sha1, md5, etc.
			if (in_array($algo, hash_algos())) {
				return hash($algo, $string, $raw);
			}
			if (function_exists($algo)) {
				return $algo($string, $raw);
			}
		}
		return false;
	}

	public static function bcrypt($string, $work_factor = 0)
	{
		if ($work_factor < 4 || $work_factor > 31) {
			$work_factor = self::BCRYPT_DEFAULT_WORK_FACTOR;
		}
		if (function_exists('password_hash')) {
			return password_hash($string, PASSWORD_BCRYPT, array('cost'=>$work_factor));
		}
		// Handle PHP < 5.3.7 security weakness
		$salt = (version_compare(PHP_VERSION, '5.3.7', '>=') ? '$2y$' : '$2a$');
		$salt .= str_pad($work_factor, 2, '0', STR_PAD_LEFT) . '$' .
			// the base64 function uses +'s and ending ='s; translate the first, and cut out the latter
			substr(str_replace('+','.',base64_encode(\Poodle\Random::bytes(16))), 0, 22)
		;
		$hash = crypt($string, $salt);
		return (strlen($hash) > 13) ? $hash : false;
	}

	public static function verify($algo, $string, $stored_hash, $raw=false)
	{
		$hash = false;
		switch ($algo)
		{
		case 'none':
		case '':
			$hash = $string;
			break;

		case 'bcrypt':
			if (function_exists('password_verify')) {
		 		return password_verify($string, $stored_hash);
			}
			$hash = crypt($string, $stored_hash);
			// Handle PHP < 5.3.7 security weakness
			if ($hash !== $stored_hash
			 && version_compare(PHP_VERSION, '5.3.7', '>=')
			 && 0 === strpos($stored_hash, '$2a$')
			) {
				$stored_hash = '$2x$' .substr($stored_hash, 4);
				$hash = crypt($string, $stored_hash);
			}
			break;

		case 'pbkdf2':
			$params = explode('$', $stored_hash);
			if (count($params) !== 4) {
				return false;
			}
			$stored_hash = base64_decode($params[3]);
			$hash = static::pbkdf2($params[0], $string, $params[2], $params[1], strlen($stored_hash), true);
			break;

		case 'scrypt':
			$params = explode('$', $stored_hash);
			if (count($params) !== 5) {
				return false;
			}
			$stored_hash = base64_decode($params[4]);
			$hash = Scrypt::hash($string, $params[3], $params[0], $params[1], $params[2], strlen($stored_hash), true);
			break;

		default: # sha1, md5, etc.
			if (in_array($algo, hash_algos())) {
				$hash = hash($algo, $string, $raw);
			} else
			if (function_exists($algo)) {
				$hash = $algo($string, $raw);
			}
		}

		return $hash === $stored_hash;
	}

	public static function hmac($algo, $data, $key, $raw=false)
	{
		if (in_array($algo, hash_algos())) {
			return hash_hmac($algo, $data, $key, $raw);
		}

		if ('sha1'==$algo && function_exists($algo))
		{
			if (mb_strlen($key,'8bit') > 64) { $key = sha1($key, true); }
			$key  = str_pad($key, 64, "\x00");
			$ipad = str_repeat(chr(0x36), 64);
			$opad = str_repeat(chr(0x5c), 64);
			$hash = sha1(($key ^ $ipad) . $data, true);
			return sha1(($key ^ $opad) . $hash, true);
		}

		return false;
	}

	public static function hmac_file($algo, $filename, $key, $raw=false)
	{
		if (in_array($algo, hash_algos())) {
			return hash_hmac_file($algo, $filename, $key, $raw);
		}
		return false;
	}

	public static function sha1($string, $raw=false) { return self::string('sha1', $string, $raw); }
	public static function md5($string,  $raw=false) { return self::string('md5',  $string, $raw); }

	// Recommended $iterations: At least 1000
	public static function pbkdf2($algo, $password, $salt, $iterations = 1000, $length = 0, $raw_output = false)
	{
		if (function_exists('hash_pbkdf2')) {
			return hash_pbkdf2($algo, $password, $salt, $iterations, $length, $raw_output);
		}

		/*
		 * PBKDF2 key derivation function as defined by RSA's PKCS #5: https://www.ietf.org/rfc/rfc2898.txt
		 *
		 * Test vectors can be found here: https://www.ietf.org/rfc/rfc6070.txt
		 *
		 * This implementation of PBKDF2 was originally created by https://defuse.ca
		 */
		$algo = strtolower($algo);
		if (!in_array($algo, hash_algos(), true)) {
			return false;
		}

		if ($iterations <= 0) {
			return false;
		}

		$hash_length = strlen(hash($algo, '', true));
		if ($length <= 0) {
			$length = $hash_length;
		}
		$block_count = ceil($length / $hash_length);

		$output = '';
		for ($i = 1; $i <= $block_count; ++$i) {
			// $i encoded as 4 bytes, big endian.
			$last = $salt . pack('N', $i);
			// first iteration
			$last = $xorsum = hash_hmac($algo, $last, $password, true);
			// perform the other $iterations - 1 iterations
			for ($j = 1; $j < $iterations; ++$j) {
				$xorsum ^= ($last = hash_hmac($algo, $last, $password, true));
			}
			$output .= $xorsum;
		}

		return $raw_output ? substr($output, 0, $length) : substr(bin2hex($output), 0, $length);
	}

}
